﻿using System;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using System.Windows;

namespace Competition.Common
{
    public enum Color
    {
        A,R,G,B,Y
    }
    public class BuildingBlock
    {
        public int X { get; set; }
        public int Y { get; set; }
        public Color Color { get; set; }
    }
    public struct Point
    {
        public int X;
        public int Y;
        public bool Vertical;

        public static bool operator == (Point pt1, Point pt2)
        {
            return pt1.X == pt2.X && pt1.Y == pt2.Y && pt1.Vertical == pt2.Vertical;
        }
        public static bool operator !=(Point pt1, Point pt2)
        {
            return !(pt1 == pt2);
        }
    }
    public class Layer
    {
        public Dictionary<Point,BuildingBlock> Blocks=new Dictionary<Point,BuildingBlock>();
    }
    public class Tower
    {
        public bool ColorDependent { get; set; }
        public bool Fill { get; set; }
        /// <summary>
        /// Načte data
        /// </summary>
        /// <param name="l"></param>
        /// <param name="blocks"></param>
        /// <returns></returns>
        public void Import(Stream s)
        {
            using (StreamReader sr = new StreamReader(s))
            {
                var hline = sr.ReadLine();
                string lasts="";
                int inf = 0;
                for (int i = 0; i <= hline.Length; i++)
                {
                    if (i < hline.Length&&char.IsLetterOrDigit(hline[i]))
                    {
                        lasts += hline[i];
                    }
                    else if(lasts.Length>0)
                    {
                        string hflag = lasts;
                        lasts = "";
                        if (inf == 0)
                        {
                            X = int.Parse(hflag);
                        }
                        else if(inf==1)
                        {
                            Y = int.Parse(hflag);
                        }
                        else if (inf == 2)
                        {
                            if (hflag.Length == 2)
                                ColorDependent = true;
                            if (hflag[0] == 'P' || hflag[0] == 'p')
                            {
                                Fill = true;
                            }
                        }
                        inf++;
                    }

                }

                while(!sr.EndOfStream)
                {
                    BuildingBlock bb = new BuildingBlock();
                    lasts = "";
                    inf = 0;
                    var line= sr.ReadLine();
                    for (int i = 0; i <= line.Length; i++)
                    {
                        if (i < line.Length&&char.IsLetterOrDigit(line[i]))
                        {
                            lasts += line[i];
                        }
                        else if (lasts.Length > 0)
                        {
                            string hflag = lasts;
                            lasts = "";
                            if (inf == 0)
                            {
                                bb.Color = (Color)Enum.Parse(typeof(Color), hflag);
                            }
                            else if (inf == 1)
                            {
                                var darr = hflag.Split('x','X').Select(x=>int.Parse(x)).OrderByDescending(x=>x).ToArray();
                                bb.X = darr[0];
                                bb.Y = darr[1];
                            }
                            else if (inf == 2)
                            {
                                int clonec = int.Parse(hflag);
                                for (int m = 0; m < clonec; m++)
                                {
                                    BuildingBlock clone = new BuildingBlock() { X = bb.X, Y = bb.Y, Color = bb.Color };
                                    BuildingBlocks.Add(clone);
                                }
                            }
                            inf++;
                        }
                    }

                }
            }
        }


        /// <summary>
        /// Načte data
        /// </summary>
        /// <param name="l"></param>
        /// <param name="blocks"></param>
        /// <returns></returns>
        public void ImportLayers(Stream s)
        {
            using (StreamReader sr = new StreamReader(s))
            {
                var hline = sr.ReadLine();
                string lasts = "";
                int inf = 0;
                for (int i = 0; i <= hline.Length; i++)
                {
                    if (i < hline.Length && char.IsLetterOrDigit(hline[i]))
                    {
                        lasts += hline[i];
                    }
                    else if (lasts.Length > 0)
                    {
                        string hflag = lasts;
                        lasts = "";
                        if (inf == 0)
                        {
                            X = int.Parse(hflag);
                        }
                        else if (inf == 1)
                        {
                            Y = int.Parse(hflag);
                        }
                        else if (inf == 2)
                        {
                            if (hflag.Length == 2)
                                ColorDependent = true;
                            if (hflag[0] == 'P' || hflag[0] == 'p')
                            {
                                Fill = true;
                            }
                        }
                        inf++;
                    }

                }
                Layer l=new Layer();
                while (!sr.EndOfStream)
                {
                    BuildingBlock bl = new BuildingBlock();
                    Point pt = new Point();
                    lasts = "";
                    inf = 0;
                    var line = sr.ReadLine();
                    for (int i = 0; i <= line.Length; i++)
                    {
                        if (i < line.Length && char.IsLetterOrDigit(line[i]))
                        {
                            lasts += line[i];
                        }
                        else if (lasts.Length > 0)
                        {
                            string hflag = lasts;
                            lasts = "";
                            if (inf == 0)
                            {
                                if (this.Layers.Count != int.Parse(hflag))
                                {
                                    l = new Layer();
                                    this.Layers.Add(l);
                                }

                            }
                            else if (inf == 1)
                            {
                                pt.X = int.Parse(hflag);
                            }
                                else if (inf == 2)
                            {
                                pt.Y = int.Parse(hflag);
                            }
                            else if (inf == 3)
                            {
                                bl.Color = (Color)Enum.Parse(typeof(Color), hflag);
                            }
                            else if (inf == 4)
                            {
                                var darr = hflag.Split('x', 'X').Select(x => int.Parse(x)).OrderByDescending(x => x).ToArray();
                                bl.X = darr[0];
                                bl.Y = darr[1];
                                l.Blocks.Add(pt, bl);
                            }
                       
                            inf++;
                        }
                    }

                }
            }
        }

        /// <summary>
        /// Uloží data
        /// </summary>
        /// <param name="l"></param>
        /// <param name="blocks"></param>
        /// <returns></returns>
        public void Export(Stream s)
        {
            using (StreamWriter sw = new StreamWriter(s))
            {
                sw.WriteLine("{0} {1} {2}", X, Y, ExportType());
                for (int i = 0; i < Layers.Count; i++)
                {
                    foreach(var block in Layers[i].Blocks)
                    {
                        bool xisx=!IsRotated;
                        if(block.Key.Vertical)
                        {
                            xisx=!xisx;
                        }
                        sw.WriteLine("{0} {1} {2} {3} {4}x{5}", i + 1, block.Key.X, block.Key.Y, block.Value.Color, xisx ? block.Value.X : block.Value.Y, xisx ? block.Value.Y : block.Value.X);
                    }
                }
            }
        }

        public string ExportType()
        {
            return (Fill ? "P" : "D") + (ColorDependent ? "B" : "");
        }

        public int X { get; set; }
        public int Y { get; set; }
        public List<BuildingBlock> BuildingBlocks = new List<BuildingBlock>();
        public List<Layer> Layers = new List<Layer>();
        /// <summary>
        /// Spočítá plán
        /// </summary>
        /// <param name="l"></param>
        /// <param name="blocks"></param>
        /// <returns></returns>
        public void SolveBlocks()
        {
            Layer l;
            while ((l = SolveLayer()) != null)
            {
                this.Layers.Add(l);
                if (BuildingBlocks.Sum(x => x.X * x.Y) < X * Y)
                    return;
            }
        }
        /// <summary>
        /// Spočítá plán s průběžným ulkádáním
        /// </summary>
        /// <param name="l"></param>
        /// <param name="blocks"></param>
        /// <returns></returns>
        public void SolveBlocksWithSaving(Stream s)
        {
            using (StreamWriter sw = new StreamWriter(s))
            {
                sw.WriteLine("{0} {1} {2}", X, Y,ExportType());
                int i=1;
                Layer l;
                Console.WriteLine("Solving layer "+ i);
                while ((l = SolveLayer()) != null)
                {
                    Console.WriteLine("Layer {0} solved", i++);
                    Console.WriteLine("Solving layer " + i);
                    foreach (var block in l.Blocks)
                    {
                        bool xisx = !IsRotated;
                        if (block.Key.Vertical)
                        {
                            xisx = !xisx;
                        }
                        sw.WriteLine("{0} {1} {2} {3} {4}x{5}", i + 1, block.Key.X, block.Key.Y, block.Value.Color, xisx ? block.Value.X : block.Value.Y, xisx ? block.Value.Y : block.Value.X);
                    }
                    if (BuildingBlocks.Sum(x => x.X * x.Y) < X * Y)
                        return;
                }
            }
        }
        /// <summary>
        /// Spočítá vrstvu
        /// </summary>
        /// <param name="l"></param>
        /// <param name="blocks"></param>
        /// <returns></returns>
        public Layer SolveLayer()
        {
            int tx = Math.Max(X, Y);
            int ty = Math.Min(X, Y);
            Layer l = new Layer();
            bool success = ColorDependent ? PlaceBlockColorDependent(l, BuildingBlocks) : PlaceBlock(l, BuildingBlocks);
            if (success)
            {
                //var sum=l.Blocks.Sum(x => x.Value.X * x.Value.Y);

                //may run faster
                foreach(var tr in l.Blocks)
                {
                    BuildingBlocks.Remove(tr.Value);
                }
                return l;
            }
            else
            {
                return null;
            }

        }

        public bool IsRotated
        {
            get
            {
                return X != Math.Max(X, Y);
            }
        }
        /// <summary>
        /// Pokusí se umístit vrstvu (volá se rekurzivně)
        /// </summary>
        /// <param name="l"></param>
        /// <param name="blocks"></param>
        /// <returns></returns>
        public bool PlaceBlock(Layer l, List<BuildingBlock> blocks)
        {
            while (true)
            {
                var list = blocks.Except(l.Blocks.Select(x => x.Value).OrderByDescending(x => x.X * x.Y));
                var fp = FreePoints(l).DefaultIfEmpty(new Point(){ X=-1}).FirstOrDefault();
                if (fp.X==-1)
                    return true;
                var fp2 = FreePoints(l).ElementAt(1);
                var ll = list.Select(x => new { s = new { X = x.X, Y = x.Y }, v = x }).GroupBy(x => x.s).OrderByDescending(x => x.Key.X*x.Key.Y);
                bool placed = false;
                foreach (var block in ll.Select(x => x.First().v).ToList())
                {
                    if (!IsOverflowing(block, fp))
                    {
                        l.Blocks.Add(fp, block);
                        if (PlaceBlock(l, blocks))
                        {
                            placed = true;
                            break;
                        }
                        else
                        {
                            l.Blocks.Remove(fp);
                        }
                    }
                    else if (!IsOverflowing(block, fp2))
                    {
                        l.Blocks.Add(fp2, block);
                        if (PlaceBlock(l, blocks))
                        {
                            placed = true;
                            break;
                        }
                        else
                        {
                            l.Blocks.Remove(fp2);
                        }
                    }
                }
                if (!placed)
                    return false;
            }

        }
        /// <summary>
        /// Pokusí se umístit vrstvu u které závisí na barvě (volá se rekurzivně)
        /// </summary>
        /// <param name="l"></param>
        /// <param name="blocks"></param>
        /// <returns></returns>
        public bool PlaceBlockColorDependent(Layer l, List<BuildingBlock> blocks)
        {
            while (true)
            {
                var list = blocks.Except(l.Blocks.Select(x => x.Value).OrderByDescending(x => x.X * x.Y));
                var fp = FreePoints(l).DefaultIfEmpty(new Point() { X = -1 }).FirstOrDefault();
                if (fp.X == -1)
                    return true;
                var fp2 = FreePoints(l).ElementAt(1);
                var ll = list.Select(x => new { s = new { X = x.X, Y = x.Y, C=x.Color }, v = x }).GroupBy(x => x.s).OrderByDescending(x => x.Key.X * x.Key.Y);
                bool placed = false;
                foreach (var block in ll.Select(x => x.First().v).ToList())
                {
                    if (!(IsBorder(fp) && FindSiblings(l, fp).Select(x => AtPoint(l, x)).Where(x => x != null).Any(x => x.Color == block.Color)))
                    {
                        if (!IsOverflowing(block, fp))
                        {

                            l.Blocks.Add(fp, block);
                            if (PlaceBlock(l, blocks))
                            {
                                placed = true;
                                break;
                            }
                            else
                            {
                                l.Blocks.Remove(fp);
                            }
                        }
                        else if (!IsOverflowing(block, fp2))
                        {
                            l.Blocks.Add(fp2, block);
                            if (PlaceBlock(l, blocks))
                            {
                                placed = true;
                                break;
                            }
                            else
                            {
                                l.Blocks.Remove(fp2);
                            }
                        }
                    }
                }
                if (!placed)
                    return false;
            }

        }
        /// <summary>
        /// Jestli bod leží na okraji
        /// </summary>
        /// <param name="l"></param>
        /// <param name="blocks"></param>
        /// <returns></returns>
        public bool IsBorder(Point pt)
        {
            int tx = Math.Max(X, Y);
            int ty = Math.Min(X, Y);
            return pt.X == 0 || pt.Y == 0 || pt.X == tx - 1 || pt.Y == ty - 1;
        }
        /// <summary>
        /// Sousední krajové body
        /// </summary>
        /// <param name="l"></param>
        /// <param name="blocks"></param>
        /// <returns></returns>
        public IEnumerable<Point> FindSiblings(Layer l,Point pt)
        {
            int tx = Math.Max(X, Y);
            int ty = Math.Min(X, Y);
            if (pt.X == 0)
            {
                if (pt.Y - 1 >= 0)
                    yield return new Point() { X = 0, Y = pt.Y - 1 };
                if (pt.Y +1 < ty)
                    yield return new Point() { X = 0, Y = pt.Y + 1 };
            }
            if (pt.Y == 0)
            {
                if (pt.X - 1 >= 0)
                    yield return new Point() { Y=0, X = pt.X - 1 };
                if (pt.X + 1 < tx)
                    yield return new Point() {Y=0, X = pt.X + 1 };
            }
            if (pt.X == tx - 1)
            {
                if (pt.Y - 1 >= 0)
                    yield return new Point() { X = tx - 1, Y = pt.Y - 1 };
                if (pt.Y + 1 < ty)
                    yield return new Point() { X = tx - 1, Y = pt.Y + 1 };
            }
            if (pt.Y == ty - 1)
            {
                if (pt.X - 1 >= 0)
                    yield return new Point() { Y = ty - 1, X = pt.X - 1 };
                if (pt.X + 1 < tx)
                    yield return new Point() { Y = ty - 1, X = pt.X + 1 };
            } 
        }
        /// <summary>
        /// Najde block na bodu
        /// </summary>
        /// <param name="l"></param>
        /// <param name="blocks"></param>
        /// <returns></returns>
        public BuildingBlock AtPoint(Layer l, Point pt)
        {
            foreach (var bl in l.Blocks)
            {
                if(bl.Key.X<=pt.X&&bl.Key.Y<=pt.Y&&(bl.Key.Vertical?bl.Value.Y:bl.Value.X)+bl.Key.X>pt.X&&(bl.Key.Vertical?bl.Value.X:bl.Value.Y)+bl.Key.Y>pt.Y)
                    return bl.Value;
            }
            return null;
        }

        /// <summary>
        /// Zbívající volná místa
        /// </summary>
        /// <param name="l"></param>
        /// <param name="blocks"></param>
        /// <returns></returns>
        public IEnumerable<Point> FreePoints(Layer l)
        {
            int tx = Math.Max(X, Y);
            int ty = Math.Min(X, Y);
            return FreePoints(tx, ty).Except(l.Blocks.SelectMany(x => UsedPoints(x.Value, x.Key)));
        }
        /// <summary>
        /// Zbívající volná místa
        /// </summary>
        /// <param name="l"></param>
        /// <param name="blocks"></param>
        /// <returns></returns>
        public IEnumerable<Point> FreePoints(Layer l,IEnumerable<BuildingBlock> blcoll)
        {
            int tx = Math.Max(X, Y);
            int ty = Math.Min(X, Y);
            return FreePoints(tx, ty).Except(l.Blocks.Where(x=>blcoll.Contains(x.Value)).SelectMany(x => UsedPoints(x.Value, x.Key)));
        }
        /// <summary>
        /// Zbívající volná místa
        /// </summary>
        /// <param name="l"></param>
        /// <param name="blocks"></param>
        /// <returns></returns>
        public IEnumerable<Point> FreePoints(int x,int y)
        {
            for (int xi = 0; xi < x; xi++)
            {
                for (int yi = 0; yi < y; yi++)
                {
                    yield return new Point() { X = xi, Y = yi, Vertical = false };
                    yield return new Point() { X = xi, Y = yi, Vertical = true };
                }
            }
        }
        /// <summary>
        /// Body použité blokem
        /// </summary>
        /// <param name="l"></param>
        /// <param name="blocks"></param>
        /// <returns></returns>
        public IEnumerable<Point> UsedPoints(BuildingBlock bl, Point ps)
        {
            for (int xi = ps.X; xi < ps.X+(ps.Vertical?bl.Y:bl.X); xi++)
            {
                for (int yi = ps.Y; yi < ps.Y + (ps.Vertical ? bl.X : bl.Y); yi++)
                {
                    yield return new Point() { X = xi, Y = yi, Vertical = false };
                    yield return new Point() { X = xi, Y = yi, Vertical = true };
                }
            }
        }
        /// <summary>
        /// Jestli block přetíká z plochy
        /// </summary>
        /// <param name="l"></param>
        /// <param name="blocks"></param>
        /// <returns></returns>
        public bool IsOverflowing(BuildingBlock bl, Point ps)
        {
            int tx = Math.Max(X, Y);
            int ty = Math.Min(X, Y);
            return !((ps.X + (ps.Vertical ? bl.Y : bl.X) <= tx) && (ps.Y + (ps.Vertical ? bl.X : bl.Y) <= ty));
        }
    }
}
